#include "edispz-controller.h"
#include "edispz-config.h"
#include "fluentd-client.h"
#include <getopt.h>

struct option_args {
  const char *config_file;
  const char *connection_list_file;
};

static struct option_args args;

static gboolean
load_config(EdispzController *controller, const gchar *filename)
{
  Bp35a1Client *client = controller->bp35a1_client;
  GKeyFile *keyfile;
  gchar *path = NULL;
  gchar *conf = NULL;
  gboolean ret = TRUE;

  keyfile = g_key_file_new ();
  if (filename == NULL)
    path = g_build_path("/", CONFIG_PATH, CONFIG_FILE, NULL);
  else
    path = g_build_path("/", filename, NULL);
    
  if (!g_key_file_load_from_file (keyfile, path, G_KEY_FILE_NONE, NULL)) {
    g_error("load config error %s", path);
    ret = FALSE;
    goto err;
  }
  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_TTY, NULL, NULL);
  if (conf == NULL) {
    g_error("\"tty=\" not found in %s", path);
    ret = FALSE;
    goto err;
  }
  bp35a1_client_set_device_name(client, conf);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_MODE, NULL, NULL);
  if (conf == NULL)
    g_warning("\"mode=\" not found in %s", path);
  else
    controller->operation_mode = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_INTERVAL_TIME, NULL, NULL);
  if (conf != NULL)
    controller->interval_time = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_OFFSET_TIME, NULL, NULL);
  if (conf != NULL)
    controller->offset_time = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_HUMAN_SENSOR_WAIT_TIME, NULL, NULL);
  if (conf != NULL)
    controller->human_sensor_wait_time = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_HEART_BEAT_TIME, NULL, NULL);
  if (conf != NULL)
    controller->heartbeat_time = (gushort)strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_DOOR_DETECT_MODE, NULL, NULL);
  if (conf != NULL)
    controller->door_detect_mode = (guchar)strtoul(conf, NULL, 10);

 err:
  g_key_file_free(keyfile);
  g_free(path);

  return ret;
}

static gint
load_connection(EdispzController *controller, const gchar *filename)
{
  Bp35a1Client *client = controller->bp35a1_client;
  gint ret;
  guint i;
  int err = 0;
  gchar *data;
  gchar **mac;
  struct stat st;
  gchar *path = NULL;

  if (filename == NULL)
    path = g_build_path("/", CONFIG_PATH, CONNECTION_FILE, NULL);
  else
    path = g_build_path("/", filename, NULL);

  g_debug("connection.list = %s",path);

  err = stat(path, &st);
  if (err){
    g_error ("load connection: Failed to stat '%s': %d", path, err);
    goto err;
  }

  if (g_file_get_contents(path, &data, NULL, NULL)) {
    mac = g_strsplit (data, "\n", -1);
    for (i = 0; mac[i]; i++) {
      if( !g_str_has_prefix(mac[i], "#")) {
	ret = bp35a1_client_whitelist_add (client, mac[i]);
	if (ret == 0)
	  g_debug ("whitelist += %s", mac[i]);
      }
    }
    g_free(mac);
    g_free(data);
  }

 err:
  g_free (path);
  return err;
}

void dump_raw_data(GByteArray *data)
{
  guint i;

  for(i=0; i<data->len; i++)
    {
      g_print("%x", data->data[i]);
    }
  g_print("\n");
}

gchar *
build_event(gchar *events, gchar *event_name, gchar *value)
{
  static gchar *fmt = "{\"event_id\":\"%s\", \"value\":\"%s\" }";
  gchar *event, *composite;

  event = g_strdup_printf(fmt, event_name, value);
  if ((strcmp("get_temperature", event_name) == 0 &&
       strcmp("-274", value) == 0) ||
      (strcmp("get_humidity", event_name) == 0 &&
       strcmp("-1", value) == 0)) {
    event = NULL;
  }

  if (events && event) {
    composite = g_strjoin(",", event, events, NULL);
    g_free(events);
    g_free(event);
  }
  else if(events) {
    composite = g_strdup(events);
    g_free(events);
  }
  else
    composite = event;

  return composite;
}

gchar *
build_json(gchar *id, gchar *events)
{
  static gchar *fmt = "{\"device_id\":\"%s\", \"data\":[%s]}";
  return g_strdup_printf(fmt, id, events);
}

void
fluentd_send_data(EdispzPacket *packet)
{
  EdispzPacketType type;
  FluentdClient *c;
  gchar *post_msg = NULL;
  gchar *events = NULL;
  gchar *mac, *mac_head, *mac_tail;

  c = fluentd_client_new (NULL);
  if (c == NULL) {
    g_warning("fail to connect fluentd server");
    return;
  }

  type = edispz_packet_get_type (packet->data);
  if (type == EDISPZ_PACKET_NORMAL ||
      type == EDISPZ_PACKET_SPECIAL ||
      type == EDISPZ_PACKET_TEST) {

    events = build_event(events, "get_door",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_DOOR_IS_OPEN));
    events = build_event(events, "get_door_count_before",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT));
    events = build_event(events, "get_door_time",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_DOOR_TIME));
    events = build_event(events, "get_temperature",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_TEMPERATURE));
    events = build_event(events, "get_humidity",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_HUMIDITY));
    events = build_event(events, "get_human_react",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_HUMAN_IS_REACT));
    events = build_event(events, "get_human_react_count",
			 edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT));

    if(type == EDISPZ_PACKET_SPECIAL ||
       type == EDISPZ_PACKET_TEST) {
      events = build_event(events, "get_battery",
			   edispz_packet_get_data (packet->data, EDISPZ_PACKET_DATA_BATTERY));
    }

    mac_head = g_strndup(packet->mac, 6);
    mac_tail = g_strndup((packet->mac) + 10, 6);
    mac = g_strconcat(mac_head, mac_tail, NULL);
    post_msg = build_json(mac, events);
    fluentd_client_post_json (c, NULL, post_msg);
    g_debug("msg = %s\n", post_msg);
    g_free(mac_head);
    g_free(mac_tail);
    g_free(mac);
    g_free(events);
    g_free(post_msg);
  }
  g_object_unref (c);
  return;
}

void dump_data(GByteArray *data)
{
  EdispzPacketType type;

  type = edispz_packet_get_type (data);
  switch (type) {
  case EDISPZ_PACKET_ACK_NACK:
    g_print ("  EDISPZ_PACKET_DATA_ACK_NACK: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_ACK_NACK));
    break;
  case EDISPZ_PACKET_NORMAL:
    g_print ("  EDISPZ_PACKET_DATA_DOOR_IS_OPEN: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_IS_OPEN));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_TIME: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_TIME));
    g_print ("  EDISPZ_PACKET_DATA_TEMPERATURE: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_TEMPERATURE));
    g_print ("  EDISPZ_PACKET_DATA_HUMIDITY: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMIDITY));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_IS_REACT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_IS_REACT));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT));
    break;
  case EDISPZ_PACKET_SPECIAL:
  case EDISPZ_PACKET_TEST:
    g_print ("  EDISPZ_PACKET_DATA_BATTERY: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_BATTERY));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_IS_OPEN: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_IS_OPEN));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_TIME: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_TIME));
    g_print ("  EDISPZ_PACKET_DATA_TEMPERATURE: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_TEMPERATURE));
    g_print ("  EDISPZ_PACKET_DATA_HUMIDITY: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMIDITY));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_IS_REACT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_IS_REACT));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT));
    g_print ("  EDISPZ_PACKET_DATA_FIRMWARE_VERSION: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_FIRMWARE_VERSION));
    g_print ("  EDISPZ_PACKET_DATA_TIMEKEEPING_FLAG: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_TIMEKEEPING_FLAG));
    break;
  default:
    g_print ("unknown\n");
    break;
  }
}

static void usage(void)
{
  g_printerr("edispz-device-app [-c CONFIG FILE PATH] [-p CONNECTION LIST PATH]\n");
  g_printerr("  -c, --config=CONFIG FILE PATH\n");
  g_printerr("      default: /etc/edispz-device-app/edispz.conf\n");
  g_printerr("  -l, --list=CONNECTION LIST PATH\n");
  g_printerr("      default: /etc/edispz-device-app/connection.list\n");
  g_printerr("  -h, --help\n");
  g_printerr("      Print usage(this message) and exit.\n");
}

static int parse_args(int argc, char *argv[]/*, char *config_file, char *connection_list_file*/)
{
  struct option longopts[] = {
    {"conifg",required_argument, NULL, 'c' },
    {"list",required_argument, NULL, 'l' },
    {"help",no_argument , NULL, 'h' },
    {0,0,0,0},
  };
  int c;

  args.config_file = NULL;
  args.connection_list_file = NULL;

  while ((c = getopt_long(argc, argv, "c:l:h", longopts, NULL)) != EOF) {
    switch (c) {
    case 'c':
      args.config_file = optarg;
      break;
    case 'l':
      args.connection_list_file = optarg;
      break;
    case 'h':
      g_printerr("h\n");
      usage();
      return 1;
    default:
      g_printerr("error\n");
      return -1;
    }
  }

  return 0;
}

int
main(int argc, char *argv[])
{
  gint ret;
  GAsyncQueue *aqueue;
  EdispzController *edispz;
  EdispzPacket *packet;

  ret = parse_args(argc, argv);
  if (ret < 0) {
    usage();
    return -1;
  }
  if (ret) {
    return 0;
  }

  aqueue = g_async_queue_new ();
  edispz = edispz_controller_new (aqueue);
  if (edispz == NULL) {
    g_print ("controller new failed.\n");
  }

  ret = load_config(edispz, args.config_file);
  if (ret != TRUE) {
    g_error("no config file");
    return -1;
  }

  ret = edispz_controller_start (edispz);
  if (ret < 0) {
    g_print ("controller start failed.\n");
  }

  ret = load_connection (edispz, args.connection_list_file);
  if (ret != 0) {
    g_error("no e-DISPZ in connection list");
    return -2;
  }

  while (TRUE) {
    packet = (EdispzPacket *) g_async_queue_pop (aqueue);
    if (packet != NULL) {
      fluentd_send_data(packet);
      delete_edispz_packet(packet);
    }
  }

  return 0;
}

